"""Module spark_tests.datetime

Returns a fixed datetime as today or now.
"""
import datetime as dt


class FakeDatetime(dt.datetime):
    """`dt.datetime` surrogate.

    Attributes:
        cls.fake_now: `date.datetime` for now,
            by default is `dt.datetime(2021, 3, 20)`.
    """
    fake_now = dt.datetime(2021, 3, 20)

    @classmethod
    def set_fake_now(cls, fake_now: dt.datetime) -> None:
        """Sets cls.fake_now."""
        cls.fake_now = fake_now

    @classmethod
    def now(cls, tz=None) -> dt.datetime:
        """Returns cls.fake_now."""
        return cls.fake_now

    @classmethod
    def today(cls) -> dt.datetime:
        """Returns cls.fake_now."""
        return cls.fake_now


class FakeDate(dt.date):
    """`dt.date` surrogate.
    """
    @classmethod
    def today(cls) -> dt.date:
        """Returns FakeDatetime.now().date()"""
        return FakeDatetime.now().date()



from pyspark.sql import SparkSession
import pyspark.sql.functions as f
import datetime as dt

bronze_table_name = "my_db.health_tracker_bronze"
silver_table_name = "my_db.health_tracker_silver"


def write_silver_table(spark: SparkSession):
    """Write rows ingested yesterday into Silver table

    * Select rows ingested yesterday from Bronze table
    * Extract and transform the raw string to columns
    * Load data into Silver table
    """
    yesterday = dt.date.today() - dt.timedelta(1)
    bronze_df = (spark.table(bronze_table_name)
                 .filter(f"p_ingestdate = '{yesterday}'"))
    json_schema = "time TIMESTAMP, name STRING, steps INTEGER"
    bronze_augmented_df = (bronze_df
                           .withColumn("nested_json",
                                       f.from_json("value", json_schema)))
    silver_df = bronze_augmented_df.select(
        f.col("nested_json.time").alias("eventtime"),
        f.col("nested_json.name").alias("name"),
        f.col("nested_json.steps").alias("steps"),
        f.col("nested_json.time").cast("DATE").alias("p_eventdate"),
    )
    (silver_df.write.format("delta").mode("append").partitionBy("p_eventdate")
     .saveAsTable(silver_table_name))



import pytest
from pytest import MonkeyPatch
from pyspark.sql import SparkSession
from pytest import FixtureRequest
import datetime as dt
import delta

from tests.fake_spark import FakeSpark
from spark_tests.datetime import FakeDatetime, FakeDate
from spark_tests.delta import FakeDeltaTable


@pytest.fixture(scope="session")
def spark_session(request: FixtureRequest) -> SparkSession:
    """Fixture for creating a `SparkSession`
    """
    spark = SparkSession.builder.getOrCreate()

    request.addfinalizer(lambda: spark.sparkContext.stop())

    return spark


@pytest.fixture(scope="session")
def fake_spark(spark_session: SparkSession) -> FakeSpark:
    return FakeSpark(spark_session)


@pytest.fixture(scope="function")
def mock_datetime(monkeypatch: MonkeyPatch) -> FakeDatetime:
    """Mocks dt.datetime
    Returns:
        FakeDatetime(2021, 3, 20)
    """
    fake_datetime = FakeDatetime(2021, 3, 20)
    fake_datetime.set_fake_now(dt.datetime(2021, 3, 20))
    monkeypatch.setattr(dt, "datetime", FakeDatetime)
    return fake_datetime


@pytest.fixture(scope="function")
def mock_date(monkeypatch: MonkeyPatch) -> FakeDate:
    """Mocks dt.date
    Returns:
        FakeDatetime(2021, 3, 20)
    """
    FakeDatetime.set_fake_now(dt.datetime(2021, 3, 20))
    monkeypatch.setattr(dt, "date", FakeDatetime)
    return FakeDate(2021, 3, 20)


@pytest.fixture(scope="function")
def mock_delta_table(monkeypatch: MonkeyPatch) -> None:
    """Mocks DeltaTable"""
    monkeypatch.setattr(delta, "DeltaTable", FakeDeltaTable)



from pyspark.sql import Row
import datetime as dt

from tests import operations
from spark_tests.sql import FakeSparkSession, FAKE_DF_WRITER
from spark_tests.datetime import FakeDate


def test_write_silver_table(fake_spark: FakeSparkSession,
                            mock_date: FakeDate) -> None:
    FAKE_DF_WRITER.clear()
    operations.write_silver_table(fake_spark)
    assert FAKE_DF_WRITER.name == "my_db.health_tracker_silver"
    assert FAKE_DF_WRITER.path is None
    assert FAKE_DF_WRITER.save_format == "delta"
    assert FAKE_DF_WRITER.save_mode == "append"
    assert FAKE_DF_WRITER.partition_by == ("p_eventdate",)
    assert FAKE_DF_WRITER.is_saved
    expected_source = [Row(eventtime=dt.datetime(2021, 3, 19, 6),
                           name="Armando Clemente", steps=1245,
                           p_eventdate=dt.date(2021, 3, 19)),
                       Row(eventtime=dt.datetime(2021, 3, 19, 6),
                           name="Meallan O'Conarain", steps=510,
                           p_eventdate=dt.date(2021, 3, 19))]
    assert FAKE_DF_WRITER.source.collect() == expected_source



"""Delta table test classes.

Merges operations will be just logged with no change to the database.
"""
from delta.tables import DeltaTable, DeltaMergeBuilder
from typing import Optional, Dict

from spark_tests.sql import FakeDataFrame, FakeSparkSession


class FakeDeltaTable(DeltaTable):
    """Logs merge operations with no change in the data.

    Attributes:
        spark: `FakeSparkSession`
        name: Optional name of the table
        path: Optional path of the table
        alias_name: alias name of the table
        source: `DataFrame` to merge
    """
    def __init__(self, spark: FakeSparkSession) -> None:
        self.spark = spark
        self.name = None
        self.path = None
        self.alias_name = None
        self.source = None

    @classmethod
    def forPath(cls, spark: FakeSparkSession, path: str) -> 'FakeDeltaTable':
        """Creates a `FakeDeltaTable` for given `path`."""
        result = cls(spark)
        result.path = path
        return result

    @classmethod
    def forName(cls, spark: FakeSparkSession, name: str) -> 'FakeDeltaTable':
        """Creates a `FakeDeltaTable` for given `name`."""
        result = cls(spark)
        result.name = name
        return result

    def alias(self, alias_name: str) -> 'FakeDeltaTable':
        """Sets alias for `self`."""
        self.alias_name = alias_name
        return self

    def merge(self, source: FakeDataFrame,
              condition: str) -> 'FakeDeltaMergeBuilder':
        """Start building `FakeDeltaMerge` from `source` `DataFrame`
        on the given merge `condition`."""
        result = FakeDeltaMergeBuilder(self, source, condition)
        return result


class FakeDeltaMerge:
    """Log of a merge operation.

    Logs merge parameters to check later in a test case,
    but does not execute the merge.

    Attributes:
        target:
            Target `DeltaTable`
        source: `DataFrame`
            Source `DataFrame`
        condition:
            Condition to match source rows with the target rows
        matched_action: {"UPDATE", "DELETE"}
            Optional action to be done on matched rows
        matched_condition:
            Optional condition to perform not_matched_action
        matched_update:
            Defines the rules of setting the values of columns
            that need to be updated.
        not_matched_action: {"INSERT"}
            Optional action to be done on not matched rows
        not_matched_condition:
            Optional condition to perform not_matched_action
        not_matched_insert:
            Defines the rules of setting the values of columns
            that need to be inserted.
        is_executed:
            Flag of execution.

    """
    def __init__(self) -> None:
        self.target: FakeDeltaTable = None
        self.source: FakeDataFrame = None
        self.condition: str = None
        self.matched_action: Optional[str] = None
        self.matched_condition: Optional[str] = None
        self.matched_update: Dict[str, str] = {}
        self.not_matched_action: Optional[str] = None
        self.not_matched_condition: Optional[str] = None
        self.not_matched_insert: Dict[str, str] = {}
        self.is_executed = False

    def clear(self) -> None:
        """Clear `self`."""
        self.__init__()


#: `FakeDeltaMerge` singleton instance.
FAKE_DELTA_MERGE = FakeDeltaMerge()


class FakeDeltaMergeBuilder(DeltaMergeBuilder):
    """Builds a `FakeDeltaMerge`.
    """
    def __init__(self, target: FakeDeltaTable, source: FakeDataFrame,
                 condition: str) -> None:
        FAKE_DELTA_MERGE.target = target
        FAKE_DELTA_MERGE.source = source
        FAKE_DELTA_MERGE.condition = condition

    def whenMatchedUpdate(
            self, condition: Optional[str] = None,
            set: Optional[Dict[str, str]] = None
    ) -> 'FakeDeltaMergeBuilder':
        FAKE_DELTA_MERGE.matched_action = "UPDATE"
        FAKE_DELTA_MERGE.matched_condition = condition
        FAKE_DELTA_MERGE.matched_update = set
        return self

    def whenMatchedDelete(
            self, condition: Optional[str] = None
    ) -> 'FakeDeltaMergeBuilder':
        FAKE_DELTA_MERGE.matched_action = "DELETE"
        FAKE_DELTA_MERGE.matched_condition = condition
        return self

    def whenNotMatchedInsert(
            self, condition: Optional[str] = None,
            values: Optional[Dict[str, str]] = None
    ) -> 'FakeDeltaMergeBuilder':
        FAKE_DELTA_MERGE.not_matched_action = "INSERT"
        FAKE_DELTA_MERGE.not_matched_condition = condition
        FAKE_DELTA_MERGE.not_matched_insert = values
        return self

    def execute(self) -> None:
        FAKE_DELTA_MERGE.is_executed = True



"""spark-tests.sql module.

Define test classes for Spark SQL:

* Queries on data frames are run regularly;

* Modification statements are just logged and does not modify data.
"""
from typing import List, Optional, Dict, Union

from pyspark import SparkContext
from pyspark.sql import SparkSession, DataFrame, DataFrameWriter, GroupedData
from pyspark.sql.types import StructType


class FakeSparkSession(SparkSession):
    """`SparkSession` proxy.

    * Queries on data frames are run regularly;

    * Modification statements are just logged and does not modify data.

    Attributes:
        real: real `SparkSession`.
        sql_queries: List of modification statements sent.
    """

    def __init__(self, real: SparkSession) -> None:
        self.real = real
        self.sql_queries: List[str] = []

    def clear(self) -> None:
        """Clear `sql_queries` list and `FAKE_DF_WRITER`"""
        self.__init__(self.real)
        FAKE_DF_WRITER.clear()

    def table(self, table_name: str) -> "FakeDataFrame":
        """Returns specified table as `FakeDataFrame`.

        Delegates to `self.real`.
        Result is returned as a `FakeDataFrame`.
        This behavior may be changed by subclasses.
        """
        return FakeDataFrame(self.real.table(table_name))

    def sql(self, sql_statement: str) -> "FakeDataFrame":
        """Logs a `sql_statement`.

        Just appends `sql_statement` into `self.sql_queries`
        with no change to data.

        Returns: empty `FakeDataFrame`.

        Args:
            sql_statement
        """
        self.sql_queries.append(sql_statement)

        return self.createDataFrame((), StructType([]))

    def createDataFrame(self, data, schema=None) -> 'FakeDataFrame':
        """Creates a `FakeDataFrame`.

        Delegates creation to `self.real`

        Returns created `DataFrame` as a `FakeDataFrame`

        """
        return FakeDataFrame(self.real.createDataFrame(data, schema))

    @property
    def sparkContext(self) -> SparkContext:
        """Returns `SparkContext`.

        Delegates to `self.real`.
        """
        return self.real.sparkContext


class FakeDataFrame(DataFrame):
    """`DataFrame` proxy.

    Attributes:
        real: real `DataFrame`
        alias_name: `DataFrame` alias name
    """

    def __init__(self, real: DataFrame) -> None:
        self.real = real
        self.alias_name: Optional[str] = None

    def __getattribute__(self, item):
        """Get attribute or method.

        if item == `write`:

            return `FAKE_DF_WRITER`

        elif item is a `DataFrame`returning method:

            delegate method execution to `self.real`

            return the result as a `FakeDataFrame`

        else:

            delegate attribute getting to `self.real`
        """

        def wrap_fake_grouped(*args, **kwargs) -> FakeGroupedData:
            return FakeGroupedData(self.real.groupBy(*args, **kwargs))

        def wrap_fake_df(*args, **kwargs) -> FakeDataFrame:
            result = FakeDataFrame(getattr(self.real, item)(*args, **kwargs))
            if item == "alias":
                result.alias_name = args[0]
            return result

        if item in {"real", "write", "alias_name"}:
            return super().__getattribute__(item)
        elif item in {"groupBy", "groupby"}:
            return wrap_fake_grouped
        elif item in {
                "agg", "alias",
                "checkpoint", "coalesce", "crossJoin",
                "distinct",
                "drop", "dropDuplicates", "drop_duplicates", "dropna",
                "exceptAll",
                "fillna", "filter",
                "intersect", "intersectAll",
                "join",
                "limit", "localCheckpoint",
                "orderBy",
                "repartition", "repartitionByRange", "replace",
                "sample", "sampleBy", "select", "selectExpr",
                "sort", "sortWithinPartitions", "subtract",
                "toDF", "transform",
                "union", "unionAll", "unionByName",
                "where", "withColumn", "withColumns", "withColumnRenamed",
                "withMetadata"}:
            return wrap_fake_df
        else:
            attrib = getattr(self.real, item)
            return attrib

    @property
    def write(self) -> "FakeDFWriter":
        """Returns FAKE_DF_WRITER.

        Set `self` as the source `DataFrame`.
        """
        FAKE_DF_WRITER.source = self
        return FAKE_DF_WRITER


class FakeGroupedData(GroupedData):
    """`GroupedData` proxy.

    Attributes:
        real: real `GroupedData`.
    """

    def __init__(self, real: GroupedData):
        self.real = real

    def agg(self, *exprs) -> FakeDataFrame:
        """Compute aggregates.

        Delegate to `self.real`

        Return result as a `FakeDataFrame`

        """
        result = FakeDataFrame(self.real.agg(*exprs))
        return result

    def pivot(
            self, pivot_col: str, values: Optional[List[str]] = None
    ) -> GroupedData:
        """Pivots a column of the current `DataFrame`.

        Delegates to `self.real`

        Return result as a `FakeGroupedData`
        """
        result = FakeGroupedData(self.real.pivot(pivot_col, values))
        return result

    def sum(self, *cols: str) -> FakeDataFrame:
        """Compute the sum for each numeric columns for each group.

        Delegates to `self.real`

        Return result as `FakeDataFrame`
        """
        result = FakeDataFrame(self.real.sum(*cols))
        return result


class FakeDFWriter(DataFrameWriter):
    """Stubs `DataFrameWriter`.

    Logs the write operation instead of actually writing the data.

    Singleton assumes that for each test case there is only one writing.

    Attributes:
        path:
            case writing to a file, the file path
        name:
            case writing to a table, the table name
        save_format:
            the format used to save
        source:
            the source `FakeDataFrame`
        save_mode:
            specifies the behavior of the save operation: "error",
            "errorifexists", "append", "overwrite", "ignore"
        partition_by:
            names of partitioning columns
        save_options:
            all other of partitioning columns
        is_saved:
            flag of saving execution
    """

    def __init__(self):
        self.path = None
        self.name = None
        self.save_format = "parquet"
        self.save_mode = "errorifexists"
        self.partition_by = ()
        self.save_options: Dict[str, str] = {}
        self.source: Optional[FakeDataFrame] = None
        self.is_saved = True

    def clear(self):
        """Clear `self` to default values.

        `self.save_format` = "parquet"

        `self.save_mode` = "errorifexists"
        """
        self.__init__()

    def format(self, format: str) -> 'FakeDFWriter':
        """Logs format."""
        self.save_format = format
        return self

    def mode(self, mode: str) -> 'FakeDFWriter':
        """Logs mode."""
        self.save_mode = mode
        return self

    def partitionBy(self, *cols: str) -> 'FakeDFWriter':
        """Logs partition columns."""
        self.partition_by = cols
        return self

    def option(self, key: str, value: str) -> 'FakeDFWriter':
        """Logs a configuration option."""
        self.save_options[key] = value
        return self

    def options(self, **options: str) -> 'FakeDFWriter':
        """Logs configuration options."""
        self.save_options = options
        return self

    def save(self, path: Optional[str] = None,
             format: Optional[str] = None,
             mode: Optional[str] = None,
             partitionBy: Optional[List[str]] = None,
             **options: str) -> None:
        """Logs current `DataFrame` rows that would be written to a file."""
        self.path = path
        self.name = None
        self.is_saved = True

    def saveAsTable(
            self,
            name: str,
            format: Optional[str] = None,
            mode: Optional[str] = None,
            partitionBy: Optional[Union[str, List[str]]] = None,
            **options: str
    ) -> None:
        """Logs current `DataFrame` rows that would be written to a table."""
        self.path = None
        self.name = name
        self.is_saved = True


#: `FakeDFWriter` singleton instance.
FAKE_DF_WRITER = FakeDFWriter()



""" Define a subclass of `FakeSparkSession` with dummy test data.
"""
import pandas as pd
from pyspark.sql.types import StructType, StructField, StringType, DateType

from spark_tests.sql import FakeSparkSession, FakeDataFrame


class FakeSpark(FakeSparkSession):
    """Proxy of class `SparkSession`

    Substitute memory data for Hive access.
    """

    def table(self, table_name: str) -> FakeDataFrame:
        """Returns test table specified by `table_name` as a `FakeDataFrame`.

.. csv-table:: my_db.health_tracker_bronze
   :file: test_db/my_db.health_tracker_bronze.csv
   :widths: auto
   :header-rows: 1

        **End of test database**

        Args:
            table_name: Nome of the table on the metastore.
        """
        if table_name == (this_table := "my_db.health_tracker_bronze"):
            pdf = pd.read_csv(f"tests/test_db/{table_name}.csv",
                              parse_dates=[1])
            schema = StructType([StructField("value", StringType()),
                                 StructField("p_ingestdate", DateType())])
            df = self.createDataFrame(pdf, schema)
            return df

        else:
            raise ValueError(f"Table not found: {table_name}")
